from django.contrib.contenttypes.models import ContentType
from functools import wraps
from inspect import getmembers
from six import reraise
from sys import exc_info

from ambition_utils.activity.models import Activity, ActivityGroup


def decorate_activity(fn, activity):
    """
    Decorator to wrap function call with activity state management
    """
    @wraps(fn)
    def decorator(*args, **kwargs):
        set_active = getattr(activity, 'active', lambda: None)
        set_active()
        try:
            ret_val = fn(*args, **kwargs)
            activity.success()
            return ret_val
        except Exception as e:
            activity.failure(str(e))
            reraise(*exc_info())
    return decorator


def track_activity(fn):
    """
    decorator to add state to method which indicates that its progress should be tracked through activities
    """
    fn.activity_enabled = True
    return fn


class ActivityManagedTaskMixin(object):
    """
    Mixin for tasks whose progress needs to be tracked through activities

    To make use of this mixin:
    1. include the mixin
    2. name the task methods that you wish track individually with "_activity" suffix
    3. provide a uuid when instantiating the task
    """
    def __init__(self, *args, **kwargs):
        if 'uuid' in kwargs:
            self.uuid = kwargs.pop('uuid', None)
            activity_names = [
                member[0] for member in getmembers(self)
                if hasattr(member[1], 'activity_enabled')
            ]
            context_object = self.get_context_object()
            context_object_id = getattr(context_object, 'id', None)
            context_object_type = None
            if context_object is not None:  # pragma: no cover
                context_object_type = ContentType.objects.get_for_model(context_object)
            self.activity_group, created = ActivityGroup.objects.get_or_create(
                uuid=self.uuid, name=self.get_activity_group_name(), context_object_id=context_object_id,
                context_object_type=context_object_type)

            # decorate the activity methods
            for activity_name in activity_names:
                activity = Activity.objects.create(name=activity_name, group=self.activity_group)
                decorated_method = decorate_activity(getattr(self, activity_name), activity)
                setattr(self, activity_name, decorated_method)

            # decorate the run method
            run_fn = decorate_activity(getattr(self, 'run'), self.activity_group)
            setattr(self, 'run', run_fn)

        super(ActivityManagedTaskMixin, self).__init__(*args, **kwargs)

    def get_activity_group_name(self):
        return self.__class__.__name__

    def get_context_object(self):
        return None



from abc import ABCMeta, abstractmethod


class OccurrenceHandler(object):
    """
    Recurrence handler class that is meant to be extended that will handle when a recurrence
    occurs for an rrule model
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def handle(self):
        """
        Handle a recurrence
        :return:
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-04-27 18:19
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('activity', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='activitygroup',
            name='context_object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='context_object_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='activity',
            name='status',
            field=models.CharField(choices=[('ACTIVE', 'ACTIVE'), ('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('SUCCESS', 'SUCCESS')], db_index=True, default='PENDING', max_length=255),
        ),
        migrations.AlterField(
            model_name='activitygroup',
            name='status',
            field=models.CharField(choices=[('ACTIVE', 'ACTIVE'), ('FAILURE', 'FAILURE'), ('SUCCESS', 'SUCCESS')], db_index=True, default='ACTIVE', max_length=255),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='steps_complete',
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name='activitygroup',
            name='steps_total',
            field=models.IntegerField(null=True),
        ),
    ]



from datetime import datetime

import pytz
from dateutil.rrule import rrule
from dateutil import parser
from django.contrib.postgres.fields import JSONField
from django.db import models, transaction
from django.utils.module_loading import import_string
from fleming import fleming
from manager_utils import bulk_update
from timezone_field import TimeZoneField


class RRuleManager(models.Manager):
    """
    Custom manager for rrule objects
    """
    def update_next_occurrences(self, rrule_objects=None):
        rrule_objects = rrule_objects or self.get_queryset()
        for rrule_object in rrule_objects:
            rrule_object.update_next_occurrence(save=False)

        bulk_update(self, rrule_objects, ['last_occurrence', 'next_occurrence'])

    @transaction.atomic
    def handle_overdue(self, **filters):
        """
        Handles any overdue rrules
        """

        # Get instances of all overdue recurrence handler classes
        instances = self.overdue_handler_class_instances(**filters)

        # Build a list of rrules that get returned from the handler
        rrules = []
        for instance in instances:
            rrules.extend(instance.handle())

        # Bulk update the next occurrences
        RRule.objects.update_next_occurrences(rrule_objects=rrules)

    def overdue_handler_class_instances(self, **filters):
        """
        Returns a set of instances for any handler with an old next_occurrence
        """

        # Get the rrule objects that are overdue and need to be handled
        rrule_objects = self.get_queryset().filter(
            next_occurrence__lte=datetime.utcnow(),
            **filters
        ).distinct(
            'occurrence_handler_path'
        )

        # Return instances of the handler classes
        return [
            rrule_object.get_occurrence_handler_class_instance()
            for rrule_object in rrule_objects
        ]


class RRule(models.Model):
    """
    Model that will hold rrule details and generate recurrences to be handled by the supplied handler
    """

    # Params used to generate the rrule
    rrule_params = JSONField()

    # Any meta data associated with the object that created this rule
    meta_data = JSONField(default=dict)

    # The timezone all dates should be converted to
    time_zone = TimeZoneField(default='UTC')

    # The last occurrence date that was handled
    last_occurrence = models.DateTimeField(null=True, default=None)

    # The next occurrence date that should be handled
    next_occurrence = models.DateTimeField(null=True, default=None)

    # A python path to the handler class used to handle when a recurrence occurs for this rrule
    # The configuration class must extend ambition_utils.rrule.handler.OccurrenceHandler
    occurrence_handler_path = models.CharField(max_length=500, blank=False, null=False)

    # Custom object manager
    objects = RRuleManager()

    def get_occurrence_handler_class_instance(self):
        """
        Gets an instance of the occurrence handler class associated with this rrule
        :rtype: ambition_utils.rrule.handler.OccurrenceHandler
        :return: The instance
        """
        return import_string(self.occurrence_handler_path)()

    def get_rrule(self):
        """
        Builds the rrule object by restoring all the params.
        The dtstart param will be converted to local time if it is set.
        :rtype: rrule
        """
        params = self.rrule_params

        # Convert next scheduled from utc back to time zone
        if params.get('dtstart') and not hasattr(params.get('dtstart'), 'date'):
            params['dtstart'] = parser.parse(params['dtstart'])

        # Convert until date from utc back to time zone
        if params.get('until') and not hasattr(params.get('until'), 'date'):
            params['until'] = parser.parse(params['until'])

        # Always cache
        params['cache'] = True

        # Return the rrule
        return rrule(**params)

    def get_next_occurrence(self, last_occurrence=None, force=False):
        """
        Builds the rrule and returns the next date in the series or None of it is the end of the series
        :param last_occurrence: The last occurrence that was generated
        :param force: If the next occurrence is none, force the rrule to generate another
        :rtype: rrule or None
        """
        # Get the last occurrence
        last_occurrence = last_occurrence or self.last_occurrence or datetime.utcnow()

        # Get the rule
        rule = self.get_rrule()

        # Convert to local time zone for getting next occurrence, otherwise time zones ahead of utc will return the same
        last_occurrence = fleming.convert_to_tz(last_occurrence, self.time_zone, return_naive=True)

        # Generate the next occurrence
        next_occurrence = rule.after(last_occurrence)

        # If next occurrence is none and force is true, force the rrule to generate another date
        if next_occurrence is None and force:
            # Keep a reference to the original rrule_params
            original_rrule_params = {}
            original_rrule_params.update(self.rrule_params)

            # Remove any limiting params
            self.rrule_params.pop('count', None)
            self.rrule_params.pop('until', None)

            # Refetch the rule
            rule = self.get_rrule()

            # Generate the next occurrence
            next_occurrence = rule.after(last_occurrence)

            # Restore the rrule params
            self.rrule_params = original_rrule_params

        # If there is a next occurrence, convert to utc
        if next_occurrence:
            next_occurrence = self.convert_to_utc(next_occurrence)

        # Return the next occurrence
        return next_occurrence

    def update_next_occurrence(self, save=True):
        """
        Sets the next_occurrence property to the next time in the series and sets the last_occurrence property
        to the previous value of next_occurrence. If the save option is True, the model will be saved. The
        save flag is typically set to False when wanting to bulk update records after updating the values
        of many models.
        :param save: Flag to save the model after updating the schedule.
        :type save: bool
        """
        if not self.next_occurrence:
            return None

        # Only handle if the current date is >= next occurrence
        if datetime.utcnow() < self.next_occurrence:
            return False

        self.last_occurrence = self.next_occurrence
        self.next_occurrence = self.get_next_occurrence(self.last_occurrence)

        # Only save if the flag is true
        if save:
            self.save(update_fields=['last_occurrence', 'next_occurrence'])

    def convert_to_utc(self, dt):
        """
        Treats the datetime object as being in the timezone of self.timezone and then converts it to utc timezone.
        :type dt: datetime
        """
        # Add timezone info
        dt = fleming.attach_tz_if_none(dt, self.time_zone)

        # Convert to utc
        dt = fleming.convert_to_tz(dt, pytz.utc, return_naive=True)

        return dt

    def save(self, *args, **kwargs):
        """
        Saves the rrule model to the database. If this is a new object, the first next_scheduled time is
        determined and set. The `dtstart` and `until` objects will be safely encoded as strings if they are
        datetime objects.
        """

        # Check if this is a new rrule object
        if self.pk is None:
            # Convert the scheduled time to utc so getting the rrule
            self.next_occurrence = self.convert_to_utc(self.rrule_params['dtstart'])

            # Get the first scheduled time according to the rrule (this converts from utc back to local time)
            self.next_occurrence = self.get_rrule()[0]

            # Convert back to utc before saving
            self.next_occurrence = self.convert_to_utc(self.next_occurrence)

        # Serialize the datetime objects if they exist
        if self.rrule_params.get('dtstart') and hasattr(self.rrule_params.get('dtstart'), 'date'):
            self.rrule_params['dtstart'] = self.rrule_params['dtstart'].strftime('%Y-%m-%d %H:%M:%S')

        if self.rrule_params.get('until') and hasattr(self.rrule_params.get('until'), 'date'):
            self.rrule_params['until'] = self.rrule_params['until'].strftime('%Y-%m-%d %H:%M:%S')

        # Call the parent save method
        super(RRule, self).save(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-28 17:08
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import timezone_field.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('rrule_params', django.contrib.postgres.fields.jsonb.JSONField()),
                ('meta_data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('time_zone', timezone_field.fields.TimeZoneField(default=b'UTC')),
                ('last_occurrence', models.DateTimeField(default=None, null=True)),
                ('next_occurrence', models.DateTimeField(default=None, null=True)),
                ('occurrence_handler_path', models.CharField(max_length=500)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-04-03 18:09
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Activity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField()),
                ('time_created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('time_updated', models.DateTimeField(auto_now=True, db_index=True)),
                ('time_finished', models.DateTimeField(db_index=True, null=True)),
                ('status', models.CharField(choices=[(b'ACTIVE', b'ACTIVE'), (b'FAILURE', b'FAILURE'), (b'PENDING', b'PENDING'), (b'SUCCESS', b'SUCCESS')], db_index=True, default=b'PENDING', max_length=255)),
                ('error_message', models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ActivityGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False, null=True)),
                ('name', models.CharField(max_length=255)),
                ('time_created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('time_finished', models.DateTimeField(db_index=True, null=True)),
                ('status', models.CharField(choices=[(b'ACTIVE', b'ACTIVE'), (b'FAILURE', b'FAILURE'), (b'SUCCESS', b'SUCCESS')], db_index=True, default=b'ACTIVE', max_length=255)),
                ('error_message', models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.AddField(
            model_name='activity',
            name='group',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='activities', to='activity.ActivityGroup'),
        ),
    ]



import datetime

import pytz
from dateutil import rrule
from django_dynamic_fixture import G
from django.test import TestCase
from freezegun import freeze_time

from ambition_utils.rrule.handler import OccurrenceHandler

from ambition_utils.rrule.models import RRule


class MockHandler(OccurrenceHandler):
    """
    Mack handler for handling an occurrence during testing
    """
    def handle(self, rrule):
        return True


class HandlerOne(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class HandlerTwo(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerTwo',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class HandlerThree(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerThree',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class RRuleManagerTest(TestCase):

    @freeze_time('1-1-2017')
    def test_run(self):
        """
        Should return the classes that have overdue rrule objects
        """

        # Make a program with HandlerOne that is not overdue
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        # Make an overdue program with HandlerOne
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        # Make an overdue program with HandlerTwo
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerTwo'
        )

        # Make a program with HandlerThree that is not overdue
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerThree'
        )

        # Get and verify the classes
        classes = {
            instance.__class__
            for instance in RRule.objects.overdue_handler_class_instances()
        }

        self.assertEqual(classes, {HandlerOne, HandlerTwo})

        # Handle overdue rrules
        RRule.objects.handle_overdue()

        # Check the recurrences. ids 2 and 3 should be updated
        recurrences = list(RRule.objects.order_by('id'))

        self.assertEqual(recurrences[0].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[1].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[2].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[3].next_occurrence, datetime.datetime(2017, 1, 2))

        # For coverage, make handler 3 overdue
        with freeze_time('1-3-2017'):
            RRule.objects.handle_overdue()


class RRuleTest(TestCase):

    def test_get_next_occurrence_first_is_occurrence(self):
        """
        First occurrence should be the dtstart
        """

        # Setup the params for creating the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
            'count': 3,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 10))
        self.assertEqual(rule.next_occurrence, None)

        # For coverage run the handler
        with freeze_time('1-3-2017'):
            handler = MockHandler()

            self.assertTrue(handler.handle(None))

    def test_get_next_occurrence_first_is_not_occurrence(self):
        """
        First occurrence should be later than the dtstart
        """

        # Create the params to create the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
            'count': 3,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert the initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 4, 1, 10))
        self.assertEqual(rule.next_occurrence, None)

    def test_get_next_occurrence_force(self):
        """
        If there is no next occurrence but we want to know what it would have been
        """
        # Create the params to create the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
            'count': 1,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert the initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, None)
        self.assertEqual(rule.get_next_occurrence(), None)
        self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2017, 3, 1, 10))

        # Coverage for returning early
        self.assertEqual(rule.update_next_occurrence(), None)

    def test_model_default_time_zone(self):
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'UTC')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))

    def test_model_different_time_zone_daily(self):
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 4, 3))

    def test_model_different_time_zone_ahead_daily(self):
        """
        Checks a time zone that is ahead of utc
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2018, 6, 19),
            'byhour': 1,
            'until': datetime.datetime(2018, 6, 20),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('Europe/London'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'Europe/London')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2018, 6, 19, 0))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2018, 6, 19, 0))
        self.assertEqual(rule.next_occurrence, None)

        # Check using a past time
        with freeze_time('1-3-2017'):
            # Use the force flag to get the next date
            self.assertEqual(rule.get_next_occurrence(), None)
            self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2018, 7, 19, 0))

        # Check using a future time
        with freeze_time('1-3-2117'):
            # Use the force flag to get the next date
            self.assertEqual(rule.get_next_occurrence(), None)
            self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2018, 7, 19, 0))

    def test_model_different_time_zone_ahead_crossing_day_daily(self):
        """
        Checks a time zone that is ahead of utc
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2018, 6, 19),
            'byhour': 0,
            'until': datetime.datetime(2018, 6, 20),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('Europe/London'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'Europe/London')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2018, 6, 18, 23))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2018, 6, 18, 23))
        self.assertEqual(rule.next_occurrence, None)

    def test_model_different_time_zone_monthly(self):
        """
        Test a monthly first day of month rule to catch case of converting tz back
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': 1,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 2, 3))

        # Cross dst
        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 2, 2))

    def test_model_different_time_zone_end_of_month(self):
        """
        Test a monthly first day of month rule to catch case of converting tz back
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': -1,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 3))

        # Test crossing dst
        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 1, 2))

    def test_model_different_time_zone_daily_with_ending_early(self):
        """
        Makes sure the time_zone is respected and ends before the ending time
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'until': datetime.datetime(2017, 1, 3, 21),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, None)

        # Save again to hit all conditions (resaving when dtstart and until are already strings)
        rule.save()

        self.assertEqual(rule.rrule_params['dtstart'], '2017-01-01 22:00:00')
        self.assertEqual(rule.rrule_params['until'], '2017-01-03 21:00:00')

    def test_model_different_time_zone_daily_with_ending_on_interval(self):
        """
        Makes sure the time_zone is respected and ends before the ending time
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'until': datetime.datetime(2017, 1, 3, 22),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 4, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 4, 3))
        self.assertEqual(rule.next_occurrence, None)

    def test_get_next_occurrence_count(self):
        """
        Makes sure the count ending is respected
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
            'count': 3,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3))
        self.assertEqual(rule.next_occurrence, None)

    @freeze_time('1-1-2016')
    def test_update_next_occurrence_ignore(self):
        """
        Make sure the next occurrence does not advance if the current time is beyond the next occurrence
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

    @freeze_time('1-1-2018')
    def test_update_next_occurrence(self):
        """
        Make sure the next occurrence does not advance if the current time is beyond the next occurrence
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))



from django.test import TestCase
from freezegun import freeze_time

from ambition_utils.time_helpers import get_time_zones


class TimeHelperTestCase(TestCase):

    @freeze_time('1-1-2017')
    def test_get_time_zones(self):
        """
        Makes sure the time zones are returned as dicts and that eastern is first
        """

        time_zones = get_time_zones()

        self.assertEqual(time_zones[0]['id'], 'US/Eastern')
        self.assertEqual(time_zones[0]['name'], 'US/Eastern (EST) (GMT -5)')
        self.assertTrue(len(time_zones) > 400)

    @freeze_time('1-1-2017')
    def test_get_time_zones_as_tuple(self):
        """
        Makes sure the time zones are returned as tuples and that eastern is first
        """

        time_zones = get_time_zones(return_as_tuple=True)

        self.assertEqual(time_zones[0][0], 'US/Eastern')
        self.assertEqual(time_zones[0][1], 'US/Eastern (EST) (GMT -5)')



from datetime import datetime
from pytz import common_timezones, timezone, exceptions as pytz_exceptions


def get_gmt_offset(tz_name, now):
    try:
        offset = timezone(tz_name).utcoffset(now)
    except pytz_exceptions.AmbiguousTimeError:  # pragma: no cover
        offset = timezone(tz_name).utcoffset(now, is_dst=False)
    except pytz_exceptions.NonExistentTimeError:  # pragma: no cover
        offset = timezone(tz_name).utcoffset(now, is_dst=False)

    offset = int(offset.total_seconds() / 3600)

    return '{0}'.format(offset) if offset < 0 else '+{0}'.format(offset)


def get_time_zones(return_as_tuple=False):
    """
    Gets timezones to display to the front end (and for validation). Orders timezones
    with common US ones first and attaches GMT offset to them.
    """
    us_tzs = [
        ('US/Eastern', 'US/Eastern (EST)'), ('US/Central', 'US/Central (CST)'),
        ('US/Mountain', 'US/Mountain (MST)'), ('US/Pacific', 'US/Pacific (PST)'),
        ('US/Arizona', 'US/Arizona'), ('US/Alaska', 'US/Alaska'), ('US/Hawaii', 'US/Hawaii')
    ]
    other_tzs = set(common_timezones) - set(us_tz[0] for us_tz in us_tzs)
    all_tzs = us_tzs + [(tz, tz) for tz in other_tzs]

    # Attach GMT values to the display names of the tzs
    now = datetime.utcnow()
    all_tzs = [
        (tz[0], '{0} (GMT {1})'.format(tz[1], get_gmt_offset(tz[0], now))) for tz in all_tzs
    ]

    return all_tzs if return_as_tuple else [{
        'id': tz[0],
        'name': tz[1],
    } for tz in all_tzs]



from django.db import models
from ambition_utils.anomaly.models import AnomalyBase
from math import floor


class FloatModel(models.Model):
    value = models.FloatField()


class IncrementalAnomaly(AnomalyBase):
    IS_INCREMENTAL = True
    name = models.CharField(max_length=100)

    def compute_uid(self):
        return self.name


class NonIncrementalAnomaly(AnomalyBase):
    IS_INCREMENTAL = False
    name = models.CharField(max_length=100)

    def compute_uid(self):
        return self.name

    def update(self):
        values = sorted(FloatModel.objects.all().values_list('value', flat=True))

        lower_index = int(floor(.01 * self.percentile_low * len(values)))
        upper_index = int(floor(.01 * self.percentile_high * len(values)))

        self.threshold_low = values[lower_index]
        self.threshold_high = values[upper_index]
        self.num_values_ingested += len(values)

        self.save()



import datetime

from dateutil import rrule
from django.test import TestCase

from ambition_utils.rrule.constants import RecurrenceEnds
from ambition_utils.rrule.forms import RecurrenceForm


class NestedRecurrenceFormTest(TestCase):
    def test_daily_never_ends(self):
        data = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': '6/1/2017',
            'byhour': '3',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
        }
        form = RecurrenceForm(data=data)
        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule[0].replace(tzinfo=None), datetime.datetime(2017, 6, 1, 3))
        self.assertEqual(rule[1].replace(tzinfo=None), datetime.datetime(2017, 6, 2, 3))
        self.assertEqual(rule[365].replace(tzinfo=None), datetime.datetime(2018, 6, 1, 3))

    def test_daily_never_ends_different_time_zone(self):
        data = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': '6/1/2017',
            'byhour': '22',
            'time_zone': 'US/Eastern',
            'ends': RecurrenceEnds.NEVER,
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rrule_model.next_occurrence, datetime.datetime(2017, 6, 2, 2))

        self.assertEqual(rule[0].replace(tzinfo=None), datetime.datetime(2017, 6, 1, 22))
        self.assertEqual(rule[1].replace(tzinfo=None), datetime.datetime(2017, 6, 2, 22))
        self.assertEqual(rule[365].replace(tzinfo=None), datetime.datetime(2018, 6, 1, 22))

        # Advance the occurrence
        rrule_model.update_next_occurrence()

        self.assertEqual(rrule_model.next_occurrence, datetime.datetime(2017, 6, 3, 2))

    def test_daily_every_4_days_3_times(self):
        data = {
            'freq': rrule.DAILY,
            'interval': 4,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.AFTER,
            'count': 3,
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule.count(), 3)

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 1))
        self.assertEqual(rule[1], datetime.datetime(2017, 6, 5))
        self.assertEqual(rule[2], datetime.datetime(2017, 6, 9))

    def test_return_if_errors(self):
        """
        Returns from the clean method if there are already errors
        """
        data = {}
        form = RecurrenceForm(data=data)

        self.assertFalse(form.is_valid())

        self.assertEqual(form.clean(), {
            'bynweekday': [],
            'byweekday': [],
            'count': None,
            'repeat_by': '',
            'until': None,
        })

    def test_missing_after_occurrences_count(self):
        data = {
            'freq': rrule.DAILY,
            'interval': 4,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.AFTER,
        }
        form = RecurrenceForm(data=data)

        self.assertFalse(form.is_valid())

    def test_daily_with_end_date(self):
        data = {
            'freq': rrule.DAILY,
            'interval': 6,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.ON,
            'until': '6/10/2017',
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule.count(), 2)

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 1))
        self.assertEqual(rule[1], datetime.datetime(2017, 6, 7))

    def test_missing_ends_on(self):
        """
        Checks that there is an end date if the recurrence is set to end on a date
        """
        data = {
            'freq': rrule.DAILY,
            'interval': 6,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.ON,
        }
        form = RecurrenceForm(data=data)

        self.assertFalse(form.is_valid())

    def test_clean_byweekday_error(self):
        """
        Makes sure invalid json is handled
        """
        form = RecurrenceForm(data={
            # invalid json
            'byweekday': '[[[eee'
        })

        self.assertEqual(form.clean_byweekday(), [])

    def test_clean_bynweekday_error(self):
        """
        Makes sure invalid json is handled
        """
        form = RecurrenceForm(data={
            # invalid json
            'bynweekday': '[[[eee'
        })

        self.assertEqual(form.clean_bynweekday(), [])

    def test_clear_ends_on_if_not_selected(self):
        """
        This handles a case with the date picker (and won't happen after date picker is fixed, but should still
        be handled) where the date field is not cleared when another ends choice is selected.
        """
        data = {
            'freq': rrule.DAILY,
            'interval': 6,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
            'until': '6/1/2017',
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['until'], '')

    def test_end_date_after_start_date(self):
        """
        Makes sure the end date is after the start date
        """
        data = {
            'freq': rrule.DAILY,
            'interval': 6,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.ON,
            'until': '5/31/2017',
        }
        form = RecurrenceForm(data=data)
        self.assertFalse(form.is_valid())

        data['until'] = '6/1/2017'
        form = RecurrenceForm(data=data)
        self.assertFalse(form.is_valid())

        data['until'] = '6/2/2017'
        form = RecurrenceForm(data=data)
        self.assertTrue(form.is_valid())

    def test_weekly_mwf_every_2_weeks_with_end(self):
        data = {
            'freq': rrule.WEEKLY,
            'interval': 2,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.ON,
            'until': '7/10/2017',
            'byweekday': '[0, 2, 4]',
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule.count(), 8)

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 2))
        self.assertEqual(rule[1], datetime.datetime(2017, 6, 12))
        self.assertEqual(rule[2], datetime.datetime(2017, 6, 14))
        self.assertEqual(rule[3], datetime.datetime(2017, 6, 16))
        self.assertEqual(rule[4], datetime.datetime(2017, 6, 26))
        self.assertEqual(rule[5], datetime.datetime(2017, 6, 28))
        self.assertEqual(rule[6], datetime.datetime(2017, 6, 30))
        self.assertEqual(rule[7], datetime.datetime(2017, 7, 10))

    def test_weekly_missing_repeat_on(self):
        data = {
            'freq': rrule.WEEKLY,
            'interval': 2,
            'dtstart': '6/1/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.ON,
            'until': '7/10/2017',
            'byweekday': '',
        }
        form = RecurrenceForm(data=data)

        self.assertFalse(form.is_valid())

    def test_monthly_missing_repeat_by(self):
        """
        Should raise a ValidationError because repeat_by is missing for monthly
        """
        data = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': '6/4/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
        }
        form = RecurrenceForm(data=data)

        self.assertFalse(form.is_valid())

    def test_monthly_every_month_day_of_month(self):
        data = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': '6/4/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
            'repeat_by': 'DAY_OF_THE_MONTH',
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 4))
        self.assertEqual(rule[1], datetime.datetime(2017, 7, 4))
        self.assertEqual(rule[2], datetime.datetime(2017, 8, 4))

    def test_monthly_every_month_day_of_week_from_start_of_month(self):
        """
        Second Monday of each month
        """
        data = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': '6/11/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
            'repeat_by': 'DAY_OF_THE_WEEK_START',
            'bynweekday': '[[0, 2]]',
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 12))
        self.assertEqual(rule[1], datetime.datetime(2017, 7, 10))
        self.assertEqual(rule[2], datetime.datetime(2017, 8, 14))

    def test_monthly_every_month_day_of_week_from_end_of_month(self):
        """
        Second Monday from end of month every other month
        """
        data = {
            'freq': rrule.MONTHLY,
            'interval': 2,
            'dtstart': '6/11/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
            'repeat_by': 'DAY_OF_THE_WEEK_END',
            'bynweekday': '[[0, -2]]'
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 19))
        self.assertEqual(rule[1], datetime.datetime(2017, 8, 21))
        self.assertEqual(rule[2], datetime.datetime(2017, 10, 23))

    def test_yearly(self):
        data = {
            'freq': rrule.YEARLY,
            'interval': 1,
            'dtstart': '6/4/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save()
        rule = rrule_model.get_rrule()

        self.assertEqual(rule[0], datetime.datetime(2017, 6, 4))
        self.assertEqual(rule[1], datetime.datetime(2018, 6, 4))
        self.assertEqual(rule[2], datetime.datetime(2019, 6, 4))

    def test_kwargs_setting(self):
        """
        Verifies that kwargs passed to the save method get set or ignored if they don't exist on the model
        """
        data = {
            'freq': rrule.YEARLY,
            'interval': 1,
            'dtstart': '6/4/2017',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
        }
        form = RecurrenceForm(data=data)

        self.assertTrue(form.is_valid())

        rrule_model = form.save(
            occurrence_handler_path='ambition_utils.rrule.handler.OccurrenceHandler',
            fake_recurrence=form
        )

        self.assertEqual(rrule_model.occurrence_handler_path, 'ambition_utils.rrule.handler.OccurrenceHandler')



from copy import deepcopy

from django.core.exceptions import ValidationError
from django.db.transaction import atomic
from django.forms.utils import ErrorDict


class NestedFormConfig(object):
    """
    Defines how a nested form is handled in the context of another form. Any form class using a subclass of
    NestedFormMixinBase can define a nested_form_configs list with instances of NestedFormConfig
    """

    def __init__(self, cls, key, required=False, field_prefix=None, required_key=None, pre=False, post=False):
        """
        Sets all default values
        :param cls: Any form class reference
        :param key: The dict key that will be used when passing around form arguments during processing.
        The value of this dict will be the return value of the save method
        :type key: str
        :param required: Determines if this form's fields are always required
        :type required: bool
        :param field_prefix: Optional field for when multiple instance of the same form are nested. This allows
        the submitted data to have a prefix before each form field name.
        :type field_prefix: str
        :param required_key: Optional form field that causes the form to be required based on its presence.
        :type required_key: bool
        :param pre: Flag to indicate that the nested form should be processed before the parent form save method. This
        allows the nested form's save value to be available in the save method arguments keyed off of the key param.
        :type pre: bool
        :param post: Flag to indicate that the nested form should be processed after the parent form save method. This
        allows the parent form's save value to be available in the post save method arguments.
        :type post: bool
        """
        self.cls = cls
        self.key = key
        self.required = required
        self.field_prefix = field_prefix
        self.required_key = required_key
        self.pre = pre
        self.post = post
        self.instance = None

        assert self.cls
        assert self.key

    def set_instance(self, *args, **kwargs):
        self.instance = self.cls(*args, **kwargs)


class NestedFormMixinBase(object):
    """
    Allows a form to contain multiple and optional nested forms. The form configurations are defined in
    nested_form_configs and is defined by instances of NestedFormConfig
    """
    nested_form_configs = []

    def __init__(self, *args, **kwargs):
        super(NestedFormMixinBase, self).__init__(*args, **kwargs)

        # Build a list of all nest form configs
        self.nested_forms = []

        # Keep track of form prefixes to guarantee multiple of the same form are properly prefixed
        form_prefixes = {}

        for nested_form_config in self.nested_form_configs:
            # Deep copy the form kwargs to pass to each form instance
            form_kwargs = deepcopy(kwargs)
            prefix = nested_form_config.field_prefix

            # Check if this form class already exists
            if nested_form_config.cls in form_prefixes:
                # Make sure both have a prefix value
                if not form_prefixes[nested_form_config.cls] or not nested_form_config.field_prefix:
                    raise ValidationError(
                        'Form {0} must have a field prefix'.format(nested_form_config.cls.__name__)
                    )

            # Set the prefix value to the form config prefix
            form_prefixes[nested_form_config.cls] = nested_form_config.field_prefix

            # Process the form field keys when there is a prefix defined on the nested form
            if form_kwargs.get('data') and prefix:
                for prefixed_key, value in deepcopy(form_kwargs['data']).items():
                    # Check if the prefix is there to replace
                    to_replace = '{0}_'.format(prefix)
                    if prefixed_key.startswith(to_replace):
                        # Replace the prefix
                        key = prefixed_key.replace(to_replace, '')
                        form_kwargs['data'][key] = value

                        # Get rid of the prefixed key
                        form_kwargs['data'].pop(prefixed_key)

            # Create the form instance and pass the form data
            nested_form_config.set_instance(*args, **form_kwargs)

            # Add the form config to the list of nested form configs
            self.nested_forms.append(nested_form_config)

    def get_pre_save_method_kwargs(self):  # pragma: no cover
        """
        Optionally return a dict of data that will be passed through the chain of save methods with
        pre-forms, parent form, and post-forms
        """
        return {}

    def get_post_save_method_kwargs(self, **kwargs):  # pragma: no cover
        """
        Optionally return a dict of data that will be passed to the post-forms. All previous form data will be
        available here including pre-save data and parent form save data.
        """
        return kwargs

    def get_required_forms(self):
        """
        Checks which forms are required based on the params and returns a list of only the required forms
        :rtype: list of NestedFormConfig
        """
        return [nested_form for nested_form in self.nested_forms if self.form_is_required(nested_form)]

    def save_form(self, **kwargs):
        """
        Hook for the parent form to save an object so that it doesn't override the mixin's save method and logic.
        """
        return None

    def form_is_required(self, nested_form):
        """
        Handles the logic to check if an individual form is required
        """
        if nested_form.required:
            return True

        # Get the required flag value
        return self.cleaned_data.get(nested_form.required_key)

    def full_clean(self):
        """
        Cleans all of self.data and populates self._errors and
        self.cleaned_data.

        This is copied from django with an addition at the bottom
        """
        # This is the django code
        self._errors = ErrorDict()
        if not self.is_bound:  # pragma: no cover
            return
        self.cleaned_data = {}
        if self.empty_permitted and not self.has_changed():  # pragma: no cover
            return

        self._clean_fields()
        self._clean_form()
        self._post_clean()

        # This is the additional code that updates the form's errors with the nested form's errors
        required_forms = self.get_required_forms()
        for form in required_forms:
            self._errors.update(form.instance.errors)


class NestedFormMixin(NestedFormMixinBase):
    @atomic
    def save(self, *args, **kwargs):
        """
        Handles getting additional arguments, calling save on pre-save forms, calling save_form on the parent form,
        and calling save on post-save forms. The form kwargs are passed through the whole chain and the parent
        form's save_form return value is returned from this method.
        """
        # Get any additional arguments that should be passed to the save methods
        form_kwargs = self.get_pre_save_method_kwargs()

        # Get all required nested forms
        required_forms = self.get_required_forms()

        # Save all pre-save forms
        for form in required_forms:
            if form.pre:
                # Store the return value keyed off the form.key property
                form_kwargs[form.key] = form.instance.save(**form_kwargs)

        # Save the parent form
        form_kwargs['form_save'] = self.save_form(**form_kwargs)

        # Get any additional post-save arguments
        form_kwargs = self.get_post_save_method_kwargs(**form_kwargs)

        # Save all post-save forms
        for form in required_forms:
            if form.post:
                # Store the return value keyed off the form.key property
                form_kwargs[form.key] = form.instance.save(**form_kwargs)

        # Return the value from the parent form's save_form method
        return form_kwargs['form_save']


class NestedModelFormMixin(NestedFormMixinBase):

    @atomic
    def save(self, commit=True):
        """
        Handles getting additional arguments, calling save on pre-save forms, calling save on the parent model form,
        and calling save on post-save forms. The form kwargs are passed through the whole chain and the parent
        form's save return value is returned from this method.
        """
        # Get any additional arguments that should be passed to the save methods
        form_kwargs = self.get_pre_save_method_kwargs()

        # Get all required nested forms
        required_forms = self.get_required_forms()

        # Save all pre-save forms
        for form in required_forms:
            if form.pre:
                # Store the return value keyed off the form.key property
                form_kwargs[form.key] = form.instance.save(**form_kwargs)

        # Save the parent form
        form_kwargs['instance'] = super(NestedModelFormMixin, self).save(commit=commit)

        # Get any additional post-save arguments
        form_kwargs = self.get_post_save_method_kwargs(**form_kwargs)

        # Save all post-save forms
        for form in required_forms:
            if form.post:
                # Store the return value keyed off the form.key property
                form_kwargs[form.key] = form.instance.save(**form_kwargs)

        # Return the value from the parent form's save method
        return form_kwargs['instance']



from django import forms
from django.core.exceptions import ValidationError
from django.test import TestCase

from ambition_utils.forms import NestedFormMixin, NestedFormMixinBase, NestedModelFormMixin, NestedFormConfig
from ambition_utils.tests.models import FakeModel


class NestedForm1(NestedFormMixin, forms.Form):
    three = forms.CharField(error_messages={
        'required': 'Three is required'
    })
    four = forms.CharField(required=False)

    def save(self, **kwargs):
        return kwargs['number'] * 10


class NestedForm2(NestedFormMixin, forms.Form):
    five = forms.CharField(error_messages={
        'required': 'Five is required'
    })
    six = forms.CharField(required=False)

    def save(self, **kwargs):
        return 2


class OptionalForm(NestedFormMixin, forms.Form):
    seven = forms.CharField(error_messages={
        'required': 'seven is required'
    })
    eight = forms.CharField(required=False)

    def save(self, **kwargs):
        return 'optional value'


class ParentForm(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=False,
            required_key='nested_form_1_required',
            pre=True,
        ),
        NestedFormConfig(
            cls=NestedForm2,
            key='nested_form_2',
            required=False,
            required_key='nested_form_2_required',
            pre=True,
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            field_prefix='optional_1',
            required_key='optional_required',
            pre=True,
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional_2',
            required=False,
            field_prefix='optional_2',
            required_key='optional_required_2',
            post=True,
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)

    # Flags for the 3 nested forms
    nested_form_1_required = forms.BooleanField(required=False)
    nested_form_2_required = forms.BooleanField(required=False)
    optional_required = forms.BooleanField(required=False)
    optional_required_2 = forms.BooleanField(required=False)

    def get_pre_save_method_kwargs(self):
        return {
            'number': 10,
            'another_number': 'ten'
        }

    def save_form(self, **kwargs):
        return 'the object'


class FormWithAlwaysRequired(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=True,
            pre=True
        ),
        NestedFormConfig(
            cls=NestedForm2,
            key='nested_form_2',
            required=True,
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=True,
            pre=True
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)

    # Flags for the 3 nested forms
    nested_form_1_required = forms.BooleanField(required=False)
    nested_form_2_required = forms.BooleanField(required=False)
    optional_required = forms.BooleanField(required=False)


class FormWithOptional(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=False,
            required_key='nested_form_1_required',
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            required_key='optional_required',
            pre=True
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)

    # Flags for the 2 nested forms
    nested_form_1_required = forms.BooleanField(required=False)
    optional_required = forms.BooleanField(required=False)


class ModelFormWithNestedForms(NestedModelFormMixin, forms.ModelForm):

    class Meta:
        model = FakeModel
        fields = ['name']

    nested_form_configs = [
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            field_prefix='optional_1',
            required_key='optional_required',
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional_2',
            required=False,
            field_prefix='optional_2',
            required_key='optional_required_2',
            post=True
        )
    ]

    optional_required = forms.BooleanField(required=False)
    optional_required_2 = forms.BooleanField(required=False)


class NestedFormMixinBaseTest(TestCase):

    def test_get_pre_save_method_kwargs(self):
        """
        Checks return value of pre save method
        """
        mixin = NestedFormMixinBase()

        self.assertEqual(mixin.get_pre_save_method_kwargs(), {})

    def test_get_post_save_method_kwargs(self):
        """
        Checks return value of post save method
        """
        mixin = NestedFormMixinBase()

        self.assertEqual(mixin.get_post_save_method_kwargs(one='two'), {'one': 'two'})

    def test_save_form(self):
        """
        Checks return value of save form method
        """
        mixin = NestedFormMixinBase()

        self.assertEqual(mixin.save_form(), None)

    def test_save(self):
        """
        Makes sure parent form save is called
        """
        class ParentForm1(forms.Form):

            def save(self):
                return 'saved'

        class ChildForm(NestedFormMixinBase, ParentForm1):
            pass

        form = ChildForm()

        self.assertEqual(form.save(), 'saved')

    def test_form_is_required(self):
        """
        Should return True when required is set, otherwise checks the flag value
        """
        class ParentForm1(NestedFormMixinBase, forms.Form):
            the_key = forms.BooleanField(required=False)

        # Required in config
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required=True
        )
        form = ParentForm1(data={})
        self.assertTrue(form.is_valid())

        self.assertTrue(form.form_is_required(nested_form))

        # Required from flag
        nested_form = nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required_key='the_key'
        )
        form = ParentForm1(data={
            'the_key': '1'
        })
        self.assertTrue(form.is_valid())

        self.assertTrue(form.form_is_required(nested_form))

        # Not required from flag
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required_key='the_key'
        )
        form = ParentForm1(data={})
        self.assertTrue(form.is_valid())

        self.assertFalse(form.form_is_required(nested_form))


class NestedFormMixinTest(TestCase):

    def test_field_prefix_validation_error(self):
        """
        Makes sure a validation error is raised if two of the same form exist but both don't have a field prefix
        """
        class Form1(forms.Form):
            pass

        class ParentForm1(NestedFormMixin, forms.Form):
            nested_form_configs = [
                NestedFormConfig(cls=Form1, key='one'),
                NestedFormConfig(cls=Form1, key='two')
            ]

        with self.assertRaises(ValidationError):
            ParentForm1()

    def test_nested_required_fields(self):
        """
        The ParentForm required field and 2 of the nested forms' required fields should be required. The second
        optional form has a field prefix and provides the required field, so it should not cause an additional
        validation error.
        """
        data = {
            'nested_form_1_required': '1',
            'nested_form_2_required': '1',
            'optional_required_2': '1',
            'optional_2_seven': 'good'
        }

        form = ParentForm(data=data)
        self.assertFalse(form.is_valid())

        self.assertEqual(len(form.errors), 3)

    def test_optional_form_flag_false(self):
        """
        Verifies an optional form is not required when its flag is false
        """
        data = {
            'nested_form_1_required': '1',
            'optional_required': 'False'
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())

        self.assertEqual(len(form.errors), 2)

    def test_optional_form_flag_missing(self):
        """
        Verifies an optional form is not required when its flag is missing
        """
        data = {
            'nested_form_1_required': '1',
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())

        self.assertEqual(len(form.errors), 2)

    def test_optional_form_flag_true(self):
        """
        Verifies an optional form is required when its flag is true
        """
        data = {
            'nested_form_1_required': 'True',
            'optional_required': 'True'
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())

        self.assertEqual(len(form.errors), 3)

    def test_full_scenario(self):
        """
        Covers all presave, save, postsave scenarios with multiple forms
        """
        data = {
            'nested_form_1_required': '1',
            'nested_form_2_required': '1',
            'optional_required_2': '1',
            'one': '1',
            'three': '3',
            'five': '5',
            'seven': '7',
        }

        form = ParentForm(data=data)

        self.assertTrue(form.is_valid())

        return_value = form.save()

        self.assertEqual(return_value, 'the object')


class NestedModelFormMixinTest(TestCase):

    def test_save(self):
        """
        Verifies the model save still works as expected
        """
        data = {
            'name': 'just a fake name',
            'optional_required': 'True',
            'optional_required_2': 'True',
            'seven': '7',
        }

        form = ModelFormWithNestedForms(data=data)

        self.assertTrue(form.is_valid())

        form.save()

        fake_model = FakeModel.objects.get()
        self.assertEqual(fake_model.name, 'just a fake name')
